class _Media:
    def __init__(self, **serie):
        self.__dict__.update(serie)
    def __repr__(self):
        return repr((self.id,self.title))
    def __str__(self):
        return f'{self.__dict__}'
    def __gt__(self, other):
        return self.id > other.id
    def __lt__(self, other):
        return self.id < other.id
    def __le__(self, other):
        return self.id <=other.id
    def __ge__(self, other):
        return self.id >=  other.id
    def __ne__(self, other):
        return self.id != other.id 
    def __eq__(self, other):
        return self.id == other.id 
    
class ErrorSearch(Exception):
    def __init__(self, **error):
        self.__dict__.update(error)
    def __str__(self):
        return f'{self.__dict__}'
    def __repr__(self):
        return repr((self.status,self.message,self.error)) 

class Anime(_Media):
    def __init__(self, **serie):
        super().__init__(**serie)
        
class AnimeListItem(_Media):
    def __init__(self, **serie):
        super().__init__(**serie)
        
class Manga(_Media):
    def __init__(self, **serie):
        super().__init__(**serie)

class MangaListItem(_Media):
    def __init__(self, **serie):
        super().__init__(**serie)

class Forum(_Media):
    def __init__(self,**serie):
        super().__init__(**serie)
    def __repr__(self):
        return repr((self.title))

class User(_Media):
    def __init__(self, **serie):
        super().__init__(**serie)
    def __repr__(self):
        return repr((self.id,self.name))
        
def set_media(media_class) -> Anime | Manga | ErrorSearch:
    def set_result(function):
        def wrapper(*args, **kwargs):
            result = function(*args, **kwargs)
            if(result[0]==200):
                return media_class(**result[1])
            else:
                result[1].update({"status": result[0]})
                return ErrorSearch(**result[1])
        return wrapper
    return set_result

def set_media_list(media_class) -> list[Anime] | list[Manga] | ErrorSearch:
    def set_result(function):
        def wrapper(*args, **kwargs):
            result = function(*args, **kwargs)
            if(result[0]==200):
                    aset = []
                    for res in result[1]["data"]:
                        try:
                            if not(res["list_status"] is None):
                                res["node"].update(res["list_status"])
                            if not(res["ranking"] is None):
                                res["node"].update(res["ranking"])
                        except:
                            None
                        aset.append(media_class(**res["node"]))
                    return aset
            else:
                result[1].update({"status": result[0]})
                return ErrorSearch(**result[1])
        return wrapper
    return set_result

def set_forum_list(function) -> list[Forum] | ErrorSearch:
    def wrapper(*args, **kwargs):
        result = function(*args, **kwargs)
        if(result[0]==200):
                aset = []
                for res in result[1]["categories"]:
                    aset.append(Forum(**res))
                return aset
        else:
            result[1].update({"status": result[0]})
            return ErrorSearch(**result[1])
    return wrapper

def set_topics_forum(function) -> list[Forum] | ErrorSearch:
    def wrapper(*args, **kwargs):
        result = function(*args, **kwargs)
        if(result[0]==200):
                aset = []
                for res in result[1]["data"]:
                    aset.append(Forum(**res))
                return aset
        else:
            result[1].update({"status": result[0]})
            return ErrorSearch(**result[1])
    return wrapper


from mal4py._utils import Auth
from mal4py._utils import get_new_code_verifier
from mal4py._utils import _MalToken as MalToken
from mal4py._utils import _MalAccount as MalAccount
from mal4py._media import Anime
from mal4py._media import Manga
from mal4py._media import Forum
from mal4py._media import User
from mal4py._media import MangaListItem
from mal4py._media import AnimeListItem
from mal4py._media import ErrorSearch


from typing import Any, Callable
import requests

_api_url = "https://api.myanimelist.net/v2/"
_secondary_api_url = "https://myanimelist.net/v1/"

class _BasicReq:
    def __init__(self, headers: dict[str, str] | None = None):
        """
        Add params from API MAL GET and Headers from query
        """
        if(headers is None):
            self.headers = {}
            self.headers["Content-Type"] = "application/x-www-form-urlencoded"            
            self.headers["X-MAL-Client-ID"] = "6114d00ca681b7701d1e15fe11a4987e"
        else:
            self.headers = headers

    def _get(self, slug: str, params: dict | None = None) -> list[int,dict]:
        """Get request to https://api.myanimelist.net/v2/"""
        resp_get = requests.get(_api_url+slug,params=params,headers=self.headers)
        return [resp_get.status_code,resp_get.json()]
    
    def _post(self, slug: str, data: dict | None = None) -> list[int,dict]:
        """Post request to https://api.myanimelist.net/v2/"""
        resp_post = requests.post(_api_url+slug,headers=self.headers,data=data)
        return [resp_post.status_code,resp_post.json()]
    
    def _post_api_v1(self, slug:str, data: dict | None = None) -> list[int,dict]:
        """Post request to https://myanimelist.net/v1/, usally for get token account"""
        resp_post = requests.post(_secondary_api_url+slug,headers=self.headers,data=data)
        return [resp_post.status_code,resp_post.json()]
    
    def _patch(self, slug:str, data: dict | None = None) -> list[int,dict]:
        """Patch request to https://api.myanimelist.net/v2/"""
        resp_patch = requests.patch(_api_url+slug,headers=self.headers,data=data)
        return [resp_patch.status_code,resp_patch.json()]
    
    def _delete(self, slug:str, data: dict | None = None) -> list[int,dict]:
        """Delete request to https://api.myanimelist.net/v2/"""
        resp_del = requests.delete(_api_url+slug,headers=self.headers)
        return [resp_del.status_code,resp_del.json()]


import json
import secrets
from mal4py._basic import _BasicReq, _secondary_api_url
from mal4py._query import MalAnime, MalManga, MalForum, MalUser

def get_new_code_verifier() -> str:
    """Generate new Code Verifier from Auth"""
    token = secrets.token_urlsafe(100)
    return token[:128]

class _MalToken():
    """Class for Generate Token for Session on API MAL"""
    def __init__(self, token_type: str,access_token: str,refresh_token: str,expires_in: int | None):
        self.token_type: str = token_type
        self.expires_in: int | None = expires_in
        self.access_token: str = access_token
        self.refresh_token: str = refresh_token
    
    def __str__(self):
        return f'{self.__dict__}'

    def _from_json_obj(obj: dict["token_type": str,"access_token": str,"refresh_token": str, "expires_in": int | None]): 
        """Get MalToken From Token JSON Object"""
        try:
            return _MalToken(
                obj["token_type"],
                obj["access_token"],
                obj["refresh_token"],
                obj["expires_in"]
            )
        except:
            raise Exception("Error: dict verify struct")

    def _from_json_string(string: str):
        """Get MalToken From Token JSON String"""
        obj: dict[
            "token_type": str,
            "access_token": str,
            "refresh_token": str,
            "expires_in": int | None,
        ] = json.loads(string)
        try:
            return _MalToken(
                obj["token_type"],
                obj["access_token"],
                obj["refresh_token"],
                obj["expires_in"]
            )
        except:
            raise Exception("Error: json_string verify struct")

    async def _from_credential(client_id: str,username: str,password: str):
        """
        **Unstable!** Get MalToken From Login And Password
        """
        DATA = {
            "client_id": client_id,
            "grant_type": "password",
            "username": username,
            "password": password
        }
        REQ = _BasicReq()
        res = REQ._post("auth/token",data=DATA)
        if(res[0]==200):
            return _MalToken(
                res[1]["token_type"],
                res[1]["access_token"],
                res[1]["refresh_token"],
                res[1]["expires_in"]
            )
        else:
            res[1].update({"status": res[0]})
            raise Exception(res[1])

    async def _from_refresh_token(client_id: str, client_secret: str, refresh_token: str):
        """Get MalToken From Refresh Token"""
        DATA = {
            "client_id": client_id,
            "client_secret": client_secret,
            "refresh_token": refresh_token,
            "grant_type": "refresh_token",
        }
        REQ = REQ = _BasicReq()
        res = REQ._post_api_v1("oauth2/token",data=DATA)
        if(res[0]==200):
            return _MalToken(
                res[1]["token_type"],
                res[1]["access_token"],
                res[1]["refresh_token"],
                res[1]["expires_in"]
            )
        else:
            res[1].update({"status": res[0]})
            raise Exception(res[1])

    async def _from_authorization_code(client_id: str,client_secret: str,code: str,code_verifier: str): 
        """
        Get _MalToken From PKCE Authorization Code
        """
        DATA = {
            "client_id": client_id,
            "client_secret": client_secret,
            "grant_type": "authorization_code",
            "code": code,
            "code_verifier": code_verifier,
        }
        REQ = REQ = _BasicReq()
        res = REQ._post_api_v1("oauth2/token",data=DATA)
        if(res[0]==200):
            return _MalToken(
                res[1]["token_type"],
                res[1]["access_token"],
                res[1]["refresh_token"],
                res[1]["expires_in"]
            )
        else:
            res[1].update({"status": res[0]})
            raise Exception(res[1])

class _MalAccount():
    """Class for create session some login methods"""
    def __init__(self, client_id: str, client_secret: str, mal_token: _MalToken | None = None):
        self.__client_id: str = client_id
        self.__client_secret: str = client_secret
        self.__mal_token: _MalToken = mal_token
        self.anime:MalAnime = MalAnime(self.__get_http_headers())
        self.manga: MalManga = MalManga(self.__get_http_headers())
        self.forum: MalForum = MalForum(self.__get_http_headers())
        self.user: MalUser = MalUser(self.__get_http_headers())
        
    @property
    def client_id(self) -> str:
        return self.__client_secret
    @property
    def client_secret(self) -> str:
        return self.__client_secret
    @property
    def mal_token(self) -> _MalToken:
        return self.__mal_token
    
    @client_id.setter
    def client_id(self, id):
        self.__client_id = id
    @client_secret.setter
    def client_secret(self, secret):
        self.__client_secret = secret
    @mal_token.setter
    def mal_token(self, mal_token: _MalToken):
        self.__mal_token = mal_token

    def stringify_token(self) -> str:
        """Get MalToken data as string"""
        if not(self.__mal_token is None):
            return json.dumps(self.__mal_token.__dict__)
        else:
            return None
    
    async def refresh_token(self):
        """Authorize with refresh Token"""
        if not(self.__mal_token is None):
            self.__mal_token = await _MalToken._from_refresh_token(
                self.__client_id,
                self.__client_secret,
                self.__mal_token
            )
        return self

    def __get_http_headers(self):
        """Header session for Request to MAL API"""
        HEADERS: dict[str, str] = {
            #'Content-Type': 'application/json',
            'Content-Type': 'application/x-www-form-urlencoded',
            "Authorization": None,
            "X-MAL-CLIENT-ID": self.__client_id
        }
        if not(self.__mal_token is None):
            HEADERS["Authorization"] = "Bearer %s" %(self.__mal_token.access_token)
        return HEADERS

class Auth():
    """Use initialized api (Auth) to login
        you must have this data:
        - client_id: your Client ID
        - client_secret: if your application has a Client Secret; OMIT in all other cases (i.e. you selected "Android", "iOS", or "Other" as App Type)
        - code: the user's Authorization Code after call get_oauth_url
        - code_verifier: the Code Verifier generated with function get_new_code_verifier()
    """
    def __init__(self,client_id: str = "6114d00ca681b7701d1e15fe11a4987e",client_secret: str = ""):
        self.__client_id = client_id
        self.__client_secret = client_secret

    def load_token(self, token: _MalToken):
        """Load MalToken saved from 'stringifyToken()'"""
        return _MalAccount(self.__client_id, self.__client_secret, token)

    def get_oauth_url(self,code_challenge: str) -> str:
        """Get OAuth url for code challenge or code verifier can use get_new_code_verifier function"""
        return  "%soauth2/authorize?response_type=code&client_id=%s&code_challenge=%s&state=RequestID42" %(_secondary_api_url,self.__client_id,code_challenge)

    async def authorize_with_refreshtoken(self,refresh_token: str) -> _MalAccount: 
        """If more time has passed you can also refresh token instead of loading last one"""
        MALTOKEN = await _MalToken._from_refresh_token(self.__client_id,self.__client_secret,refresh_token)
        return _MalAccount(self.__client_id,self.__client_secret,MALTOKEN)

    async def authorize_with_code(self, code: str,code_challenge: str) -> _MalAccount:
        """
        Open returned url, accept oauth and use returned code to authorize
        It is actually a `code_verifier` but mal accepts code_challenge here instead
        """
        MALTOKEN = await _MalToken._from_authorization_code(self.__client_id,self.__client_secret,code,code_challenge)
        return _MalAccount(self.__client_id, self.__client_secret, MALTOKEN)

    def authorize_with_json_obj(self, json_obj: dict["token_type": str,"access_token": str,"refresh_token": str, "expires_in": int | None]) -> _MalAccount:
        """Get MalToken From Token JSON Dict"""
        MALTOKEN = _MalToken._from_json_obj(json_obj)
        return _MalAccount(self.__client_id, self.__client_secret, MALTOKEN)
    
    def authorize_with_json_string(self, json_str: str) -> _MalAccount:
        """Get MalToken From Token JSON String"""
        MALTOKEN = _MalToken._from_json_string(json_str)
        return _MalAccount(self.__client_id,self.__client_secret, MALTOKEN)
    
    async def unstable_login(self,username: str, password: str) -> _MalAccount:
        """### Login to API using login and password `(Unstable!)`"""
        MALTOKEN = await _MalToken._from_credential(self.__client_id,username,password)
        return _MalAccount(self.__client_id,self.__client_secret, MALTOKEN)
    
    async def guest_login(self) -> _MalAccount:
        """### You can use some API functions, BEWARE `(Unstable!)`"""
        return _MalAccount(self.__client_id,self.__client_secret, None)


from mal4py._basic import _BasicReq
from mal4py._media import set_media, set_media_list, set_forum_list, Anime, Manga, Forum, User, AnimeListItem, MangaListItem, ErrorSearch

class MalAnime():
    def __init__(self, header_session: dict[str,str]):
        self.__slug_type = "anime"
        self.__HEADERS = header_session
        self.__query = _BasicReq(self.__HEADERS)

    @set_media_list(Anime)
    def get_list(self,q:str,limit:int=100,offset:int=0,fields:str="id,title,main_picture") -> list[Anime] | ErrorSearch:
        """Get Anime List by name"""
        payload =  {"q" : q[0:63], "limit": limit, "offset": offset, "fields": fields}
        return self.__query._get(self.__slug_type,payload)
    
    @set_media(Anime)
    def get_details(self,anime_id: int,fields: str="id,title,main_picture") -> Anime | ErrorSearch:
        """Get Anime by ID MAL"""
        payload =  {"fields":fields}
        return self.__query._get(self.__slug_type+"/%i"%(anime_id),payload)

    @set_media_list(Anime)
    def get_seasonal(self,year: int,season:str,sort:str="anime_score",limit:int=100,offset:int=0,fields:str="id,title,main_picture")-> list[Anime] | ErrorSearch:
        """Get a List Anime Seasonal actually"""
        payload =  {"sort":sort,"limit": limit, "offset": offset, "fields": fields}
        return self.__query._get(self.__slug_type+"/season/%i/%s"%(year,season),payload)

    @set_media_list(Anime)
    def get_suggested(self,limit:int=100,offset:int=0,fields:str="id,title,main_picture") -> list[Anime] | ErrorSearch:
        """Get a List of Anime suggested"""
        payload =  {"limit": limit, "offset": offset, "fields": fields}
        return self.__query._get(self.__slug_type+"/suggestions",payload)

    @set_media_list(Anime)
    def get_ranking(self,ranking_type:str = "all",limit:int=100,offset:int=0,fields:str="id,title,main_picture") -> list[Anime] | ErrorSearch:
        """Get a List Ranking Anime!"""
        payload =  {"ranking_type":ranking_type,"limit": limit, "offset": offset, "fields": fields}
        return self.__query._get(self.__slug_type+"/ranking",payload)
    
class MalManga():
    def __init__(self, header_session: dict[str,str]):
        self.__slug_type = "manga"
        self.__HEADERS = header_session
        self.__query = _BasicReq(self.__HEADERS)

    @set_media_list(Manga)
    def get_list(self,q:str,limit:int=100,offset:int=0,fields:str="id,title,main_picture") -> list[Manga] | ErrorSearch:
        """Get Manga List by name"""
        payload =  {"q" : q[0:64], "limit": limit, "offset": offset, "fields": fields}
        return self.__query._get(self.__slug_type,payload)
    
    @set_media(Manga)
    def get_details(self,manga_id:int,fields: str="id,title,main_picture") -> Manga | ErrorSearch:
        """Get Manga by ID MAL"""
        payload =  {"fields":fields}
        return self.__query._get(self.__slug_type+"/%i"%(manga_id),payload)
    
    @set_media_list(Manga)
    def get_ranking(self,ranking_type:str = "all",limit:int=100,offset:int=0,fields:str="id,title,main_picture") -> list[Manga] | ErrorSearch:
        """Get a List Ranking Manga!"""
        payload =  {"ranking_type":ranking_type,"limit": limit, "offset": offset, "fields": fields}
        return self.__query._get(self.__slug_type+"/ranking",payload)
    
class MalForum():
    
    def __init__(self, header_session: dict[str,str]):
        self.__slug_type = "forum"
        self.__HEADERS = header_session
        self.__query = _BasicReq(self.__HEADERS)
        
    @set_forum_list
    def get_board(self) -> list[Forum] | ErrorSearch:
        """Get forum boards"""
        payload =  {}
        return self.__query._get(self.__slug_type+"/boards",payload)
    
    @set_media(Forum)
    def get_topic_detail(self,topic_id:int,limit:int=100,offset:int=0) -> list[Forum] | ErrorSearch:
        """Get forum topic details"""
        payload =  {"limit": limit, "offset": offset}
        return self.__query._get(self.__slug_type+"/topic/%s" %(topic_id),payload)
    
    def get_topics(self,board_id:int,subboard_id:int,q:str,topic_user_name:str="",user_name:str="",limit:int=100,offset:int=0) -> list:
        """Get forum topic details - Unstable this function return JSON DATA maybe your repair struct!"""
        payload =  {"q":q[0:64],"board_id":board_id,"subboard_id":subboard_id,"topic_user_name":topic_user_name,"user_name":user_name,"limit": limit, "offset": offset,"sort":"recent"}
        return self.__query._get(self.__slug_type+"/topics",payload)
    
class MalUser():
    def __init__(self, header_session: dict[str,str]):
        self.__slug_type = "users"
        self.__HEADERS = header_session
        self.__query = _BasicReq(self.__HEADERS)
    
    @set_media(User)
    def get_my_info(self,user_id:str="@me",fields:str="id,name,anime_statistics,is_supporter") -> User | ErrorSearch:
        """Get my user information"""
        payload =  {"fields":fields}
        return self.__query._get(self.__slug_type+"/%s" %(user_id),payload)
    
    @set_media(MangaListItem)
    def update_mangalist_status(self,manga_id,body:dict) -> MangaListItem | ErrorSearch:
        """Add specified manga to my manga list.
            If specified manga already exists, update its status.
            This endpoint updates only values specified by the parameter."""
        return self.__query._patch("manga/%s/my_list_status" %(manga_id),data=body)
    
    def delete_mangalist_item(self,manga_id) -> list:
        """If the specified manga does not exist in user's manga list, this endpoint does nothing and returns 404 Not Found."""
        return self.__query._delete("manga/%s/my_list_status" %(manga_id))[0]
    
    @set_media_list(MangaListItem)
    def get_mangalist(self,user_id:str="@me",fields:str="list_status",status="",sort:str="list_updated_at",limit:int=100,offset:int=0) -> list[MangaListItem] | ErrorSearch:
        """Get Manga list of User"""
        payload =  {"fields":fields,"status":status,"sort":sort,"limit":limit,"offset":offset}
        return self.__query._get(self.__slug_type+"/%s/mangalist" %(user_id),payload)
    
    @set_media(AnimeListItem)
    def update_animelist_status(self,anime_id,body:dict[str,str]) -> AnimeListItem | ErrorSearch:
        """Add specified anime to my anime list.
            If specified anime already exists, update its status.
            This endpoint updates only values specified by the parameter."""
        return self.__query._patch("anime/%s/my_list_status" %(anime_id),data=body)
    
    def delete_animelist_item(self,anime_id) -> list:
        """If the specified anime does not exist in user's anime list, this endpoint does nothing and returns 404 Not Found."""
        return self.__query._delete("anime/%s/my_list_status" %(anime_id))[0]
    
    @set_media_list(AnimeListItem)
    def get_animelist(self,user_id:str="@me",fields:str="list_status",status="",sort:str="list_updated_at",limit:int=100,offset:int=0) -> list[AnimeListItem] | ErrorSearch:
        """Get Manga list of User"""
        payload =  {"fields":fields,"status":status,"sort":sort,"limit":limit,"offset":offset}
        return self.__query._get(self.__slug_type+"/%s/animelist" %(user_id),payload)